import asyncio
from abc import ABCMeta, abstractmethod
from typing import (
  Dict,
  List,
  Optional,
  Sequence,
  Tuple,
  TypeVar,
  Union,
)

from pylabrobot.io.usb import USB
from pylabrobot.liquid_handling.backends.backend import (
  LiquidHandlerBackend,
)
from pylabrobot.liquid_handling.backends.tecan.errors import (
  TecanError,
  error_code_to_exception,
)
from pylabrobot.liquid_handling.liquid_classes.tecan import (
  TecanLiquidClass,
  get_liquid_class,
)
from pylabrobot.liquid_handling.standard import (
  Drop,
  DropTipRack,
  MultiHeadAspirationContainer,
  MultiHeadAspirationPlate,
  MultiHeadDispenseContainer,
  MultiHeadDispensePlate,
  Pickup,
  PickupTipRack,
  ResourceDrop,
  ResourceMove,
  ResourcePickup,
  SingleChannelAspiration,
  SingleChannelDispense,
)
from pylabrobot.resources import (
  Coordinate,
  Liquid,
  Resource,
  TecanPlate,
  TecanPlateCarrier,
  TecanTip,
  TecanTipRack,
  Tip,
  TipSpot,
  Trash,
)

T = TypeVar("T")


class TecanLiquidHandler(LiquidHandlerBackend, metaclass=ABCMeta):
  """
  Abstract base class for Tecan liquid handling robot backends.
  """

  @abstractmethod
  def __init__(
    self,
    packet_read_timeout: int = 120,
    read_timeout: int = 300,
    write_timeout: int = 300,
  ):
    """

    Args:
      packet_read_timeout: The timeout for reading packets from the Tecan machine in seconds.
      read_timeout: The timeout for reading from the Tecan machine in seconds.
    """

    super().__init__()
    self.io = USB(
      human_readable_device_name="Tecan EVO",
      packet_read_timeout=packet_read_timeout,
      read_timeout=read_timeout,
      write_timeout=write_timeout,
      id_vendor=0x0C47,
      id_product=0x4000,
    )

    self._cache: Dict[str, List[Optional[int]]] = {}

  def _assemble_command(self, module: str, command: str, params: List[Optional[int]]) -> str:
    """Assemble a firmware command to the Tecan machine.

    Args:
      module: 2 character module identifier (C5 for LiHa, ...)
      command: 3 character command identifier
      params: list of integer parameters

    Returns:
      A string containing the assembled command.
    """

    cmd = module + command + ",".join(str(a) if a is not None else "" for a in params)
    return f"\02{cmd}\00"

  def parse_response(self, resp: bytes) -> Dict[str, Union[str, int, List[Union[int, str]]]]:
    """Parse a machine response string

    Args:
      resp: The response string to parse.

    Returns:
      A dictionary containing the parsed values.
    """

    s = resp.decode("utf-8", "ignore")
    module = s[1:3]
    ret = int(resp[3]) ^ (1 << 7)
    if ret != 0:
      raise error_code_to_exception(module, ret)

    # below line changed to revie int and str.
    data: List[Union[int, str]] = []
    for x in s[3:-1].split(","):
      if len(x) == 0:  # Skip empty values
        continue
      # RoMa(C1) and LiHa(C5) should only have integer values
      # MCA can return both integers and strings
      data.append(int(x) if x.lstrip("-").isdigit() else x)

    return {"module": module, "data": data}

  async def send_command(
    self,
    module: str,
    command: str,
    params: Optional[List[Optional[int]]] = None,
    write_timeout: Optional[int] = None,
    read_timeout: Optional[int] = None,
    wait=True,
  ):
    """Send a firmware command to the Tecan machine. Caches `set` commands and ignores if
    redundant.

    Args:
      module: 2 character module identifier (C5 for LiHa, ...)
      command: 3 character command identifier
      params: list of integer parameters
      write_timeout: write timeout in seconds. If None, `self.write_timeout` is used.
      read_timeout: read timeout in seconds. If None, `self.read_timeout` is used.
      wait: If True, wait for a response. If False, return `None` immediately after sending.

    Returns:
      A dictionary containing the parsed response, or None if no response was read within `timeout`.
    """

    if command[0] == "S" and params is not None:
      k = module + command
      if k in self._cache and self._cache[k] == params:
        return
      self._cache[k] = params

    cmd = self._assemble_command(module, command, [] if params is None else params)

    await self.io.write(cmd.encode(), timeout=write_timeout)
    if not wait:
      return None

    resp = await self.io.read(timeout=read_timeout)
    return self.parse_response(resp)

  async def setup(self):
    await super().setup()
    await self.io.setup()

  async def stop(self):
    await self.io.stop()


class EVOBackend(TecanLiquidHandler):
  """
  Interface for the Tecan Freedom EVO series
  """

  LIHA = "C5"
  ROMA = "C1"
  MCA = "W1"
  PNP = "W2"

  def __init__(
    self,
    diti_count: int = 0,
    packet_read_timeout: int = 12,
    read_timeout: int = 60,
    write_timeout: int = 60,
  ):
    """Create a new EVO interface.

    Args:
      packet_read_timeout: timeout in seconds for reading a single packet.
      read_timeout: timeout in seconds for reading a full response.
      write_timeout: timeout in seconds for writing a command.
    """

    super().__init__(
      packet_read_timeout=packet_read_timeout,
      read_timeout=read_timeout,
      write_timeout=write_timeout,
    )

    self._num_channels: Optional[int] = None
    self.diti_count = diti_count
    # channels [num_channels - diti_count, num_channels) configured for disposable tips

    self._liha_connected: Optional[bool] = None
    self._roma_connected: Optional[bool] = None
    self._pnp_connected: Optional[bool] = None
    self._mca_connected: Optional[bool] = None

    self._z_traversal_height = 210  # mm, the default value for SHZ command
    self._z_roma_traversal_height = (
      68.7  # mm, is what was used to develop this but possibly too low
    )

  @property
  def num_channels(self) -> int:
    """The number of pipette channels present on the robot."""

    if self._num_channels is None:
      raise RuntimeError("has not loaded num_channels, forgot to call `setup`?")
    return self._num_channels

  @property
  def liha_connected(self) -> bool:
    """Whether LiHa arm is present on the robot."""

    if self._liha_connected is None:
      raise RuntimeError("liha_connected not set, forgot to call `setup`?")
    return self._liha_connected

  @property
  def roma_connected(self) -> bool:
    """Whether RoMa arm is present on the robot."""

    if self._roma_connected is None:
      raise RuntimeError("roma_connected not set, forgot to call `setup`?")
    return self._roma_connected

  @property
  def pnp_connected(self) -> bool:
    """Whether PnP arm is present on the robot."""

    if self._pnp_connected is None:
      raise RuntimeError("pnp_connected not set, forgot to call `setup`?")
    return self._pnp_connected

  @property
  def mca_connected(self) -> bool:
    """Whether MCA arm is present on the robot."""

    if self._mca_connected is None:
      raise RuntimeError("mca_connected not set, forgot to call `setup`?")
    return self._mca_connected

  def serialize(self) -> dict:
    return {**super().serialize(), **self.io.serialize()}

  async def setup(self):
    """Setup

    Creates a USB connection and finds read/write interfaces.
    """

    await super().setup()

    self._liha_connected = await self.setup_arm(EVOBackend.LIHA)
    self._mca_connected = await self.setup_arm(EVOBackend.MCA)
    self._roma_connected = await self.setup_arm(EVOBackend.ROMA)

    if self.roma_connected:  # position_initialization_x in reverse order from setup_arm
      self.roma = RoMa(self, EVOBackend.ROMA)
      await self.roma.position_initialization_x()
      # move to home position (TBD) after initialization
      await self._park_roma()
    if self.mca_connected:
      self.mca = Mca(self, EVO.MCA)
      # await self.mca.position_initialization_x() # function does not work for mca.
      await self._park_mca()

    if self.liha_connected:
      self.liha = LiHa(self, EVOBackend.LIHA)
      await self.liha.position_initialization_x()

    self._num_channels = await self.liha.report_number_tips()
    self._x_range = await self.liha.report_x_param(5)
    self._y_range = (await self.liha.report_y_param(5))[0]
    self._z_range = (await self.liha.report_z_param(5))[0]

    # Initialize plungers. Assumes wash station assigned at rail 1.
    await self.liha.set_z_travel_height([self._z_range] * self.num_channels)
    await self.liha.position_absolute_all_axis(45, 1031, 90, [1200] * self.num_channels)
    await self.liha.initialize_plunger(self._bin_use_channels(list(range(self.num_channels))))
    await self.liha.position_valve_logical([1] * self.num_channels)
    await self.liha.move_plunger_relative([100] * self.num_channels)
    await self.liha.position_valve_logical([0] * self.num_channels)
    await self.liha.set_end_speed_plunger([1800] * self.num_channels)
    await self.liha.move_plunger_relative([-100] * self.num_channels)
    await self.liha.position_absolute_all_axis(45, 1031, 90, [self._z_range] * self.num_channels)

  async def setup_arm(self, module):
    try:
      if module == EVO.MCA:
        await self.send_command(module, command="PIB")

      await self.send_command(module, command="PIA")
    except TecanError as e:
      if e.error_code == 5:
        return False
      raise e

    if module != EVO.MCA:
      await self.send_command(module, command="BMX", params=[2])

    return True

  async def _park_liha(self):
    await self.liha.set_z_travel_height([self._z_range] * self.num_channels)
    await self.liha.position_absolute_all_axis(45, 1031, 90, [self._z_range] * self.num_channels)

  async def _park_roma(self):
    await self.roma.set_vector_coordinate_position(1, 9000, 2000, 2464, 1800, None, 1, 0)
    await self.roma.action_move_vector_coordinate_position()

  async def _park_mca(self):
    """Moves the MCA arm to a safe park position to prevent collision."""
    # TODO CHANGE TO USE CORRECT FUNCTIONS

    # Ensure MCA is initialized before moving
    await self.send_command(EVO.MCA, command="PIA")
    await asyncio.sleep(0.5)

    # Raise MCA Z-axis first to avoid collision
    await self.send_command(EVO.MCA, command="PAA", params=[None, None, 2000])  # Raise Z-axis
    await asyncio.sleep(1)

    # Move MCA to parking position (adjust X, Y as needed)
    await self.send_command(EVO.MCA, command="PAA", params=[6000, 1000, None])
    await asyncio.sleep(1)

    # Stop movement to prevent drifting
    await self.send_command(EVO.MCA, command="BMA", params=[0, 0, 0])
    await asyncio.sleep(0.5)

  # ============== LiquidHandlerBackend methods ==============

  async def aspirate(
    self, ops: List[SingleChannelAspiration], use_channels: List[int]
  ):  # TODO: pass in operation parameters to override TecanLiquidClass defaults
    """Aspirate liquid from the specified channels.

    Args:
      ops: The aspiration operations to perform.
      use_channels: The channels to use for the operations.
    """

    # Get positions including offsets
    x_positions, y_positions, z_positions = self._liha_positions(ops, use_channels)

    tecan_liquid_classes = [
      get_liquid_class(
        target_volume=op.volume,
        liquid_class=Liquid.WATER,
        tip_type=op.tip.tip_type,
      )
      if isinstance(op.tip, TecanTip)
      else None
      for op in ops
    ]

    ys = int(ops[0].resource.get_absolute_size_y() * 10)
    zadd: List[Optional[int]] = [0] * self.num_channels
    for i, channel in enumerate(use_channels):
      par = ops[i].resource.parent
      if par is None:
        continue
      if not isinstance(par, TecanPlate):
        raise ValueError(f"Operation is not supported by resource {par}.")
      # TODO: calculate defaults when area is not specified
      zadd[channel] = round(ops[i].volume / par.area * 10)

    # moves such that first channel is over first location
    x, _ = self._first_valid(x_positions)
    y, yi = self._first_valid(y_positions)
    assert x is not None and y is not None
    await self.liha.set_z_travel_height([self._z_range] * self.num_channels)
    await self.liha.position_absolute_all_axis(
      x,
      y - yi * ys,
      ys,
      [z if z else self._z_range for z in z_positions["travel"]],
    )
    # TODO check channel positions match resource positions

    # aspirate airgap
    pvl, sep, ppr = self._aspirate_airgap(use_channels, tecan_liquid_classes, "lag")
    if any(ppr):
      await self.liha.position_valve_logical(pvl)
      await self.liha.set_end_speed_plunger(sep)
      await self.liha.move_plunger_relative(ppr)

    # perform liquid level detection
    # TODO: verify for other liquid detection modes
    if any(tlc.aspirate_lld if tlc is not None else None for tlc in tecan_liquid_classes):
      tlc, _ = self._first_valid(tecan_liquid_classes)
      assert tlc is not None
      detproc = tlc.lld_mode  # must be same for all channels?
      sense = tlc.lld_conductivity
      await self.liha.set_detection_mode(detproc, sense)
      ssl, sdl, sbl = self._liquid_detection(use_channels, tecan_liquid_classes)
      await self.liha.set_search_speed(ssl)
      await self.liha.set_search_retract_distance(sdl)
      await self.liha.set_search_z_start(z_positions["start"])
      await self.liha.set_search_z_max(list(z if z else self._z_range for z in z_positions["max"]))
      await self.liha.set_search_submerge(sbl)
      shz = [min(z for z in z_positions["travel"] if z)] * self.num_channels  # TODO: max?
      await self.liha.set_z_travel_height(shz)
      await self.liha.move_detect_liquid(self._bin_use_channels(use_channels), zadd)
      await self.liha.set_z_travel_height([self._z_range] * self.num_channels)

    # aspirate + retract
    # SSZ: z_add / (vol / asp_speed)
    zadd = [min(z, 32) if z else None for z in zadd]
    ssz, sep, stz, mtr, ssz_r = self._aspirate_action(ops, use_channels, tecan_liquid_classes, zadd)
    await self.liha.set_slow_speed_z(ssz)
    await self.liha.set_end_speed_plunger(sep)
    await self.liha.set_tracking_distance_z(stz)
    await self.liha.move_tracking_relative(mtr)
    await self.liha.set_slow_speed_z(ssz_r)
    await self.liha.move_absolute_z(z_positions["start"])  # TODO: use retract_position and offset

    # aspirate airgap
    pvl, sep, ppr = self._aspirate_airgap(use_channels, tecan_liquid_classes, "tag")
    await self.liha.position_valve_logical(pvl)
    await self.liha.set_end_speed_plunger(sep)
    await self.liha.move_plunger_relative(ppr)

  async def dispense(self, ops: List[SingleChannelDispense], use_channels: List[int]):
    """Dispense liquid from the specified channels.

    Args:
      ops: The dispense operations to perform.
      use_channels: The channels to use for the dispense operations.
    """

    x_positions, y_positions, z_positions = self._liha_positions(ops, use_channels)
    ys = int(ops[0].resource.get_absolute_size_y() * 10)

    tecan_liquid_classes = [
      get_liquid_class(
        target_volume=op.volume,
        liquid_class=Liquid.WATER,
        tip_type=op.tip.tip_type,
      )
      if isinstance(op.tip, TecanTip)
      else None
      for op in ops
    ]

    x, _ = self._first_valid(x_positions)
    y, yi = self._first_valid(y_positions)
    assert x is not None and y is not None
    await self.liha.set_z_travel_height([z if z else self._z_range for z in z_positions["travel"]])
    await self.liha.position_absolute_all_axis(
      x,
      y - yi * ys,
      ys,
      [z if z else self._z_range for z in z_positions["dispense"]],
    )

    sep, spp, stz, mtr = self._dispense_action(ops, use_channels, tecan_liquid_classes)
    await self.liha.set_end_speed_plunger(sep)
    await self.liha.set_stop_speed_plunger(spp)
    await self.liha.set_tracking_distance_z(stz)
    await self.liha.move_tracking_relative(mtr)

  async def pick_up_tips(self, ops: List[Pickup], use_channels: List[int]):
    """Pick up tips from a resource.

    Args:
      ops: The pickup operations to perform.
      use_channels: The channels to use for the pickup operations.
    """

    assert min(use_channels) >= self.num_channels - self.diti_count, (
      f"DiTis can only be configured for the last {self.diti_count} channels"
    )

    # Get positions including offsets
    x_positions, y_positions, z_positions = self._liha_positions(ops, use_channels)

    # move channels
    ys = int(ops[0].resource.get_absolute_size_y() * 10)
    x, _ = self._first_valid(x_positions)
    y, yi = self._first_valid(y_positions)
    assert x is not None and y is not None
    await self.liha.set_z_travel_height([self._z_range] * self.num_channels)
    await self.liha.position_absolute_all_axis(
      x, y - yi * ys, ys, [self._z_range] * self.num_channels
    )

    # aspirate airgap
    pvl: List[Optional[int]] = [None] * self.num_channels
    sep: List[Optional[int]] = [None] * self.num_channels
    ppr: List[Optional[int]] = [None] * self.num_channels
    for channel in use_channels:
      pvl[channel] = 0
      sep[channel] = 70 * 6  # ? 12, always 70?
      ppr[channel] = 10 * 3  # ? 6
    await self.liha.position_valve_logical(pvl)
    await self.liha.set_end_speed_plunger(sep)
    await self.liha.move_plunger_relative(ppr)

    # get tips
    first_z_start, _ = self._first_valid(z_positions["start"])
    assert first_z_start is not None, "Could not find a valid z_start position"
    await self.liha.get_disposable_tip(
      self._bin_use_channels(use_channels), first_z_start - 227, 210
    )

  async def drop_tips(self, ops: List[Drop], use_channels: List[int]):
    """Drops tips.

    Args:
      ops: The drop operations to perform.
      use_channels: The channels to use for the drop operations.
    """

    assert min(use_channels) >= self.num_channels - self.diti_count, (
      f"DiTis can only be configured for the last {self.diti_count} channels"
    )
    assert all(isinstance(op.resource, (Trash, TipSpot)) for op in ops), (
      "Must drop in waste container or tip rack"
    )

    # Get positions including offsets
    x_positions, y_positions, _ = self._liha_positions(ops, use_channels)

    # move channels
    ys = int(ops[0].resource.get_absolute_size_y() * 10)  # was 90
    x, _ = self._first_valid(x_positions)
    y, yi = self._first_valid(y_positions)
    assert x is not None and y is not None
    await self.liha.set_z_travel_height([self._z_range] * self.num_channels)
    await self.liha.position_absolute_all_axis(
      x,
      y - yi * ys,
      ys,
      [self._z_range] * self.num_channels,
    )

    # TODO check channel positions match resource positions for z-axis
    await self.liha._drop_disposable_tip(self._bin_use_channels(use_channels), discard_height=0)

  async def pick_up_tips96(self, pickup: PickupTipRack):
    raise NotImplementedError("MCA not implemented yet")

  async def drop_tips96(self, drop: DropTipRack):
    raise NotImplementedError("MCA not implemented yet")

  async def aspirate96(
    self, aspiration: Union[MultiHeadAspirationPlate, MultiHeadAspirationContainer]
  ):
    raise NotImplementedError("MCA not implemented yet")

  async def dispense96(self, dispense: Union[MultiHeadDispensePlate, MultiHeadDispenseContainer]):
    raise NotImplementedError("MCA not implemented yet")

  async def pick_up_resource(self, pickup: ResourcePickup):
    # TODO: implement PnP for moving tubes
    assert self.roma_connected

    z_range = await self.roma.report_z_param(5)
    x, y, z = self._roma_positions(
      pickup.resource, pickup.resource.get_location_wrt(self.deck), z_range
    )
    h = int(pickup.resource.get_absolute_size_y() * 10)

    # move to resource
    await self.roma.set_smooth_move_x(1)
    await self.roma.set_fast_speed_x(10000)
    await self.roma.set_fast_speed_y(5000, 1500)
    await self.roma.set_fast_speed_z(1300)
    await self.roma.set_fast_speed_r(5000, 1500)
    await self.roma.set_vector_coordinate_position(1, x, y, z["safe"], 900, None, 1, 0)
    await self.roma.action_move_vector_coordinate_position()
    await self.roma.set_smooth_move_x(0)

    # pick up resource
    await self.roma.position_absolute_g(900)  # TODO: verify
    await self.roma.set_target_window_class(1, 0, 0, 0, 135, 0)
    await self.roma.set_vector_coordinate_position(1, x, y, z["travel"], 900, None, 1, 1)
    # TODO verify z param
    await self.roma.set_vector_coordinate_position(1, x, y, z["end"], 900, None, 1, 0)
    await self.roma.action_move_vector_coordinate_position()
    await self.roma.set_fast_speed_y(3500, 1000)
    await self.roma.set_fast_speed_r(2000, 600)
    await self.roma.set_gripper_params(100, 75)
    await self.roma.grip_plate(h - 100)

  async def move_picked_up_resource(self, move: ResourceMove):
    raise NotImplementedError()

  async def drop_resource(self, drop: ResourceDrop):
    """Drop a resource like a plate or a lid using the integrated robotic arm."""

    z_range = await self.roma.report_z_param(5)
    x, y, z = self._roma_positions(
      drop.resource, drop.resource.get_location_wrt(self.deck), z_range
    )
    xt, yt, zt = self._roma_positions(drop.resource, drop.destination, z_range)

    # move to target
    await self.roma.set_target_window_class(1, 0, 0, 0, 135, 0)
    await self.roma.set_target_window_class(2, 0, 0, 0, 53, 0)
    await self.roma.set_target_window_class(3, 0, 0, 0, 55, 0)
    await self.roma.set_target_window_class(4, 45, 0, 0, 0, 0)
    await self.roma.set_vector_coordinate_position(1, x, y, z["end"], 900, None, 1, 1)
    await self.roma.set_vector_coordinate_position(2, x, y, z["travel"], 900, None, 1, 2)
    await self.roma.set_vector_coordinate_position(3, x, y, z["safe"], 900, None, 1, 3)
    await self.roma.set_vector_coordinate_position(4, xt, yt, zt["safe"], 900, None, 1, 4)
    await self.roma.set_vector_coordinate_position(5, xt, yt, zt["travel"], 900, None, 1, 3)
    await self.roma.set_vector_coordinate_position(6, xt, yt, zt["end"], 900, None, 1, 0)
    await self.roma.action_move_vector_coordinate_position()

    # release resource
    await self.roma.position_absolute_g(900)
    await self.roma.set_fast_speed_y(5000, 1500)
    await self.roma.set_fast_speed_r(5000, 1500)
    await self.roma.set_vector_coordinate_position(1, xt, yt, zt["end"], 900, None, 1, 1)
    await self.roma.set_vector_coordinate_position(2, xt, yt, zt["travel"], 900, None, 1, 2)
    await self.roma.set_vector_coordinate_position(3, xt, yt, zt["safe"], 900, None, 1, 0)
    await self.roma.action_move_vector_coordinate_position()
    await self.roma.set_fast_speed_y(3500, 1000)
    await self.roma.set_fast_speed_r(2000, 600)

  def can_pick_up_tip(self, channel_idx: int, tip: Tip) -> bool:
    if not isinstance(tip, TecanTip):
      return False
    return True

  def _first_valid(self, lst: List[Optional[T]]) -> Tuple[Optional[T], int]:
    """Returns first item in list that is not None"""

    for i, v in enumerate(lst):
      if v is not None:
        return v, i
    return None, -1

  def _bin_use_channels(self, use_channels: List[int]) -> int:
    """Converts use_channels to a binary coded tip representation."""

    b = 0
    for channel in use_channels:
      b += 1 << channel
    return b

  def _liha_positions(
    self,
    ops: Sequence[Union[SingleChannelAspiration, SingleChannelDispense, Pickup, Drop]],
    use_channels: List[int],
  ) -> Tuple[
    List[Optional[int]],
    List[Optional[int]],
    Dict[str, List[Optional[int]]],
  ]:
    """Creates lists of x, y, and z positions used by LiHa ops."""

    x_positions: List[Optional[int]] = [None] * self.num_channels
    y_positions: List[Optional[int]] = [None] * self.num_channels
    z_positions: Dict[str, List[Optional[int]]] = {
      "travel": [None] * self.num_channels,
      "start": [None] * self.num_channels,
      "dispense": [None] * self.num_channels,
      "max": [None] * self.num_channels,
    }

    def get_z_position(z, z_off, tip_length):
      # TODO: simplify z units
      return int(self._z_range - z + z_off * 10 + tip_length)  # TODO: verify z formula

    for i, (op, channel) in enumerate(zip(ops, use_channels)):
      location = ops[i].resource.get_location_wrt(self.deck) + op.resource.center()
      x_positions[channel] = int((location.x - 100 + op.offset.x) * 10)
      y_positions[channel] = int((346.5 - location.y + op.offset.y) * 10)  # TODO: verify

      par = ops[i].resource.parent
      if not isinstance(par, (TecanPlate, TecanTipRack)):
        raise ValueError(f"Operation is not supported by resource {par}.")
      # TODO: calculate defaults when z-attribs are not specified
      tip_length = int(ops[i].tip.total_tip_length * 10)
      # z travel seems to only be used for aspiration and dispense right now
      if isinstance(op, (SingleChannelAspiration, SingleChannelDispense)):
        z_positions["travel"][channel] = round(self._z_traversal_height * 10)
      z_positions["start"][channel] = get_z_position(
        par.z_start, par.get_location_wrt(self.deck).z + op.offset.z, tip_length
      )
      z_positions["dispense"][channel] = get_z_position(
        par.z_dispense, par.get_location_wrt(self.deck).z + op.offset.z, tip_length
      )
      z_positions["max"][channel] = get_z_position(
        par.z_max, par.get_location_wrt(self.deck).z + op.offset.z, tip_length
      )

    return x_positions, y_positions, z_positions

  def _aspirate_airgap(
    self,
    use_channels: List[int],
    tecan_liquid_classes: List[Optional[TecanLiquidClass]],
    airgap: str,
  ) -> Tuple[List[Optional[int]], List[Optional[int]], List[Optional[int]]]:
    """Creates parameters used to aspirate airgaps.

    Args:
      airgap: `lag` for leading airgap, `tag` for trailing airgap.

    Returns:
      pvl: position_valve_logical
      sep: set_end_speed_plunger
      ppr: move_plunger_relative
    """

    pvl: List[Optional[int]] = [None] * self.num_channels
    sep: List[Optional[int]] = [None] * self.num_channels
    ppr: List[Optional[int]] = [None] * self.num_channels

    for i, channel in enumerate(use_channels):
      tlc = tecan_liquid_classes[i]
      assert tlc is not None
      pvl[channel] = 0
      if airgap == "lag":
        sep[channel] = int(
          tlc.aspirate_lag_speed * 6
        )  # 6? TODO: verify step unit (half step per second)
        ppr[channel] = int(tlc.aspirate_lag_volume * 3)  # 3? (Relative position in full steps)
      elif airgap == "tag":
        sep[channel] = int(tlc.aspirate_tag_speed * 6)  # 6?
        ppr[channel] = int(tlc.aspirate_tag_volume * 3)  # 3?

    return pvl, sep, ppr

  def _liquid_detection(
    self,
    use_channels: List[int],
    tecan_liquid_classes: List[Optional[TecanLiquidClass]],
  ) -> Tuple[List[Optional[int]], List[Optional[int]], List[Optional[int]]]:
    """Creates parameters use for liquid detection.

    Returns:
      ssl: set_search_speed
      sdl: set_search_retract_distance
      sbl: set_search_submerge
    """

    ssl: List[Optional[int]] = [None] * self.num_channels
    sdl: List[Optional[int]] = [None] * self.num_channels
    sbl: List[Optional[int]] = [None] * self.num_channels

    for i, channel in enumerate(use_channels):
      tlc = tecan_liquid_classes[i]
      assert tlc is not None
      ssl[channel] = int(tlc.lld_speed * 10)
      sdl[channel] = int(tlc.lld_distance * 10)
      sbl[channel] = int(tlc.aspirate_lld_offset * 10)

    return ssl, sdl, sbl

  def _aspirate_action(
    self,
    ops: Sequence[Union[SingleChannelAspiration, SingleChannelDispense]],
    use_channels: List[int],
    tecan_liquid_classes: List[Optional[TecanLiquidClass]],
    zadd: List[Optional[int]],
  ) -> Tuple[
    List[Optional[int]],
    List[Optional[int]],
    List[Optional[int]],
    List[Optional[int]],
    List[Optional[int]],
  ]:
    """Creates parameters used for aspiration action.

    Args:
      zadd: distance moved while aspirating

    Returns:
      ssz: set_slow_speed_z
      sep: set_end_speed_plunger
      stz: set_tracking_distance_z
      mtr: move_tracking_relative
      ssz_r: set_slow_speed_z retract
    """

    ssz: List[Optional[int]] = [None] * self.num_channels
    sep: List[Optional[int]] = [None] * self.num_channels
    stz: List[Optional[int]] = [-z if z else None for z in zadd]  # TODO: verify max cutoff
    mtr: List[Optional[int]] = [None] * self.num_channels
    ssz_r: List[Optional[int]] = [None] * self.num_channels

    for i, channel in enumerate(use_channels):
      tlc = tecan_liquid_classes[i]
      z = zadd[channel]
      assert tlc is not None and z is not None
      flow_rate = ops[i].flow_rate or tlc.aspirate_speed
      sep[channel] = int(flow_rate * 6)  # 6?
      ssz[channel] = round(z * flow_rate / ops[i].volume)
      volume = tlc.compute_corrected_volume(ops[i].volume)
      mtr[channel] = round(volume * 3)  # 3?  # Relative position in full steps
      ssz_r[channel] = int(tlc.aspirate_retract_speed * 10)

    return ssz, sep, stz, mtr, ssz_r

  def _dispense_action(
    self,
    ops: Sequence[Union[SingleChannelAspiration, SingleChannelDispense]],
    use_channels: List[int],
    tecan_liquid_classes: List[Optional[TecanLiquidClass]],
  ) -> Tuple[
    List[Optional[int]],
    List[Optional[int]],
    List[Optional[int]],
    List[Optional[int]],
  ]:
    """Creates parameters used for dispense action.

    Returns:
      sep: set_end_speed_plunger
      spp: set_stop_speed_plunger
      stz: set_tracking_distance_z
      mtr: move_tracking_relative
    """

    sep: List[Optional[int]] = [None] * self.num_channels
    spp: List[Optional[int]] = [None] * self.num_channels
    stz: List[Optional[int]] = [None] * self.num_channels
    mtr: List[Optional[int]] = [None] * self.num_channels

    for i, channel in enumerate(use_channels):
      tlc = tecan_liquid_classes[i]
      assert tlc is not None
      flow_rate = ops[i].flow_rate or tlc.dispense_speed
      sep[channel] = int(flow_rate * 6)  # 6?
      spp[channel] = int(tlc.dispense_breakoff * 6)  # 6? half step per second
      stz[channel] = 0
      volume = (
        tlc.compute_corrected_volume(ops[i].volume)
        + tlc.aspirate_lag_volume
        + tlc.aspirate_tag_volume
      )
      mtr[channel] = -round(volume * 3)  # 3?

    return sep, spp, stz, mtr

  def _roma_positions(
    self, resource: Resource, offset: Coordinate, z_range: int
  ) -> Tuple[int, int, Dict[str, int]]:
    """Creates x, y, and z positions used by RoMa ops."""

    parent = resource.parent  # PlateHolder
    if parent is None:
      raise ValueError(f"Operation is not supported by resource {resource}.")
    parent = parent.parent  # PlateCarrier
    # TODO: this is probably the current plate carrier, not the destination.
    # Also, we should just support any coordinate as the destination.
    if not isinstance(parent, TecanPlateCarrier):
      raise ValueError(f"Operation is not supported by resource {parent}.")

    if (
      parent.roma_x is None
      or parent.roma_y is None
      or parent.roma_z_safe is None
      or parent.roma_z_end is None
    ):
      raise ValueError(f"Operation is not supported by resource {parent}.")
    x_position = int((offset.x - 100) * 10 + parent.roma_x)
    y_position = int((347.1 - (offset.y + resource.get_absolute_size_y())) * 10 + parent.roma_y)
    z_positions = {
      "safe": z_range - int(parent.roma_z_safe),
      "travel": int(self._z_roma_traversal_height * 10),
      "end": z_range - int(parent.roma_z_end - offset.z * 10),
    }

    return x_position, y_position, z_positions


class EVOArm:
  """
  Provides firmware commands for EVO arms. Caches arm positions.
  """

  _pos_cache: Dict[str, int] = {}

  def __init__(self, backend: EVOBackend, module: str):
    self.backend = backend
    self.module = module

  async def position_initialization_x(self):
    """Reinitializes X-axis of the arm."""

    await self.backend.send_command(module=self.module, command="PIX")

  async def report_x_param(self, param: int) -> int:
    """Report current parameter for x-axis.

    Args:
      param: 0 - current position, 5 - actual machine range
    """

    resp: List[int] = (
      await self.backend.send_command(module=self.module, command="RPX", params=[param])
    )["data"]
    return resp[0]

  async def report_y_param(self, param: int) -> List[int]:
    """Report current parameters for y-axis.

    Args:
      param: 0 - current position, 5 - actual machine range
    """

    resp: List[int] = (
      await self.backend.send_command(module=self.module, command="RPY", params=[param])
    )["data"]
    return resp


class LiHa(EVOArm):
  """
  Provides firmware commands for the LiHa
  """

  async def initialize_plunger(self, tips):
    """Initializes plunger and valve drive

    Args:
      tips: binary coded tip select
    """
    await self.backend.send_command(module=self.module, command="PID", params=[tips])

  async def report_z_param(self, param: int) -> List[int]:
    """Report current parameters for z-axis.

    Param:
    0   Report current position in 1/10 mm.
    1   Report acceleration in 1/10 mm/s².
    2   Report fast speed in 1/10 mm/s.
    3   Report initialization speed in 1/10 mm.
    4   Report initialization offset in 1/10 mm.
    5   Report actual machine range in 1/10 mm.
    6   Report deviation in encoder increments.
    7   Report every time 0.
    8   Report scale adjust factor.
    9   Report slow speed in 1/10 mm/s.
    10  Report axis scale factor.
    11  Report target position in 1/10 mm.
    12  Report travel position in 1/10 mm.
    """

    resp: List[int] = (
      await self.backend.send_command(module=self.module, command="RPZ", params=[param])
    )["data"]
    return resp

  async def report_number_tips(self) -> int:
    """Report number of tips on arm."""

    resp: List[int] = (
      await self.backend.send_command(module=self.module, command="RNT", params=[1])
    )["data"]
    return resp[0]

  async def position_absolute_all_axis(self, x: int, y: int, ys: int, z: List[int]):
    """Position absolute for all LiHa axes.

    Args:
      x: absolute x position in 1/10 mm, must be in allowed machine range
      y: absolute y position in 1/10 mm, must be in allowed machine range
      ys: absolute y spacing in 1/10 mm, must be between 90 and 380
      z: absolute z position in 1/10 mm for each channel, must be in
         allowed machine range

    Raises:
      TecanError: if moving to the target position causes a collision
    """

    cur_x = EVOArm._pos_cache.setdefault(self.module, await self.report_x_param(0))
    for module, pos in EVOArm._pos_cache.items():
      if module == self.module:
        continue
      if cur_x < x and cur_x < pos < x:  # moving right
        raise TecanError("Invalid command (collision)", self.module, 2)
      if cur_x > x and cur_x > pos > x:
        raise TecanError("Invalid command (collision)", self.module, 2)
      if abs(pos - x) < 1500:
        raise TecanError("Invalid command (collision)", self.module, 2)

    await self.backend.send_command(module=self.module, command="PAA", params=list([x, y, ys] + z))

    EVOArm._pos_cache[self.module] = x

  async def position_valve_logical(self, param: List[Optional[int]]):
    """Position valve logical for each channel.

    Args:
      param: 0 - outlet, 1 - inlet, 2 - bypass
    """

    await self.backend.send_command(module=self.module, command="PVL", params=param)

  async def set_end_speed_plunger(self, speed: List[Optional[int]]):
    """Set end speed for plungers.

    Args:
      speed: speed for each plunger in half step per second, must be between
             5 and 6000
    """

    await self.backend.send_command(module=self.module, command="SEP", params=speed)

  async def move_plunger_relative(self, rel: List[Optional[int]]):
    """Move plunger relative upwards (dispense) or downards (aspirate).

    Args:
      rel: relative position for each plunger in full steps, must be between
           -3150 and 3150
    """

    await self.backend.send_command(module=self.module, command="PPR", params=rel)

  async def set_detection_mode(self, proc: int, sense: int):
    """Set liquid detection mode.

    Args:
      proc: detection procedure (7 for double detection sequential with
            retract and extra submerge)
      sense: conductivity (1 for high)
    """

    await self.backend.send_command(module=self.module, command="SDM", params=[proc, sense])

  async def set_search_speed(self, speed: List[Optional[int]]):
    """Set search speed for liquid search commands.

    Args:
      speed: speed for each channel in 1/10 mm/s, must be between 1 and 1500
    """

    await self.backend.send_command(module=self.module, command="SSL", params=speed)

  async def set_search_retract_distance(self, dist: List[Optional[int]]):
    """Set z-axis retract distance for liquid search commands.

    Args:
      dist: retract distance for each channel in 1/10 mm, must be in allowed
            machine range
    """

    await self.backend.send_command(module=self.module, command="SDL", params=dist)

  async def set_search_submerge(self, dist: List[Optional[int]]):
    """Set submerge for liquid search commands.

    Args:
      dist: submerge distance for each channel in 1/10 mm, must be between
            -1000 and max z range
    """

    await self.backend.send_command(module=self.module, command="SBL", params=dist)

  async def set_search_z_start(self, z: List[Optional[int]]):
    """Set z-start for liquid search commands.

    Args:
      z: start height for each channel in 1/10 mm, must be in allowed machine range
    """

    await self.backend.send_command(module=self.module, command="STL", params=z)

  async def set_search_z_max(self, z: List[Optional[int]]):
    """Set z-max for liquid search commands.

    Args:
      z: max for each channel in 1/10 mm, must be in allowed machine range
    """

    await self.backend.send_command(module=self.module, command="SML", params=z)

  async def set_z_travel_height(self, z):
    """Set z-travel height.

    Args:
      z: travel heights in absolute 1/10 mm for each channel, must be in allowed
         machine range + 20
    """
    await self.backend.send_command(module=self.module, command="SHZ", params=z)

  async def move_detect_liquid(self, channels: int, zadd: List[Optional[int]]):
    """Move tip, detect liquid, submerge.

    Args:
      channels: binary coded tip select
      zadd: required distance to travel downwards in 1/10 mm for each channel,
            must be between 0 and z-start - z-max
    """

    await self.backend.send_command(
      module=self.module,
      command="MDT",
      params=[channels] + [None] * 3 + zadd,
    )

  async def set_slow_speed_z(self, speed: List[Optional[int]]):
    """Set slow speed for z.

    Args:
      speed: speed in 1/10 mm/s for each channel, must be between 1 and 4000
    """

    await self.backend.send_command(module=self.module, command="SSZ", params=speed)

  async def set_tracking_distance_z(self, rel: List[Optional[int]]):
    """Set z-axis relative tracking distance used by dispense and aspirate.

    Args:
      rel: relative value in 1/10 mm for each channel, must be between
            -2100 and 2100
    """

    await self.backend.send_command(module=self.module, command="STZ", params=rel)

  async def move_tracking_relative(self, rel: List[Optional[int]]):
    """Move tracking relative. Starts the z-drives and dilutors simultaneously
        to achieve a synchronous tracking movement.

    Args:
      rel: relative position for each plunger in full steps, must be between
           -3150 and 3150
    """

    await self.backend.send_command(module=self.module, command="MTR", params=rel)

  async def move_absolute_z(self, z: List[Optional[int]]):
    """Position absolute with slow speed z-axis

    Args:
      z: absolute position in 1/10 mm for each channel, must be in
         allowed machine range
    """

    await self.backend.send_command(module=self.module, command="MAZ", params=z)

  async def set_stop_speed_plunger(self, speed: List[Optional[int]]):
    """Set stop speed for plungers

    Args:
      speed: speed for each plunger in half step per second, must be between
             50 and 2700
    """

    await self.backend.send_command(module=self.module, command="SPP", params=speed)

  async def get_disposable_tip(self, tips, z_start, z_search):
    """Picks up tips

    Args:
      tips: binary coded tip select
      z_start: position in 1/10 mm where searching begins
      z_search: search distance in 1/10 mm, range within a tip must be found
    """

    await self.backend.send_command(
      module=self.module,
      command="AGT",
      params=[tips, z_start, z_search, 0],
    )

  async def discard_disposable_tip_high(self, tips):
    """Drops tips
    Discards at the Z-axes initialization height
    Args:
      tips: binary coded tip select
    """

    await self.backend.send_command(module=self.module, command="ADT", params=[tips])

  async def _drop_disposable_tip(self, tips, discard_height):
    """Drops tips
    Discards at a variable Z-axis initialization height

    Args:
      tips: binary coded tip select
      discard_height: binary. 0 above tip rack, 1 in tip rack
    """

    await self.backend.send_command(
      module=self.module, command="AST", params=[tips, discard_height]
    )


class Mca(EVOArm):
  pass


class RoMa(EVOArm):
  """
  Provides firmware commands for the RoMa plate robot
  """

  async def report_z_param(self, param: int) -> int:
    """Report current parameter for z-axis.

    Args:
      param: 0 - current position, 5 - actual machine range
    """

    resp: List[int] = (
      await self.backend.send_command(module=self.module, command="RPZ", params=[param])
    )["data"]
    return resp[0]

  async def report_r_param(self, param: int) -> int:
    """Report current parameter for r-axis.

    Args:
      param: 0 - current position, 5 - actual machine range
    """

    resp: List[int] = (
      await self.backend.send_command(module=self.module, command="RPR", params=[param])
    )["data"]
    return resp[0]

  async def report_g_param(self, param: int) -> int:
    """Report current parameter for g-axis.

    Args:
      param: 0 - current position, 5 - actual machine range
    """

    resp: List[int] = (
      await self.backend.send_command(module=self.module, command="RPG", params=[param])
    )["data"]
    return resp[0]

  async def set_smooth_move_x(self, mode: int):
    """Sets X-axis smooth move.

    Args:
      mode: 0 - active, 1 - normal acceleration and speed used
    """

    await self.backend.send_command(module=self.module, command="SSM", params=[mode])

  async def set_fast_speed_x(self, speed: Optional[int], accel: Optional[int] = None):
    """Set fast speed and acceleration for X-axis.

    Args:
      speed: fast speed in 1/10 mm/s
      accel: acceleration in 1/10 mm/s^2
    """

    await self.backend.send_command(module=self.module, command="SFX", params=[speed, accel])

  async def set_fast_speed_y(self, speed: Optional[int], accel: Optional[int] = None):
    """Set fast speed and acceleration for Y-axis.

    Args:
      speed: fast speed in 1/10 mm/s
      accel: acceleration in 1/10 mm/s^2
    """

    await self.backend.send_command(module=self.module, command="SFY", params=[speed, accel])

  async def set_fast_speed_z(self, speed: Optional[int], accel: Optional[int] = None):
    """Set fast speed and acceleration for Z-axis.

    Args:
      speed: fast speed in 1/10 mm/s
      accel: acceleration in 1/10 mm/s^2
    """

    await self.backend.send_command(module=self.module, command="SFZ", params=[speed, accel])

  async def set_fast_speed_r(self, speed: Optional[int], accel: Optional[int] = None):
    """Set fast speed and acceleration for R-axis.

    Args:
      speed: fast speed in 1/10 dg/s
      accel: acceleration in 1/10 dg/s^2
    """

    await self.backend.send_command(module=self.module, command="SFR", params=[speed, accel])

  async def set_vector_coordinate_position(
    self,
    v: int,
    x: int,
    y: int,
    z: int,
    r: int,
    g: Optional[int],
    speed: int,
    tw: int = 0,
  ):
    """Sets vector coordinate positions into table.

    Args:
      v: vector to be defined, must be between 1 and 100
      x: absolute x position in 1/10 mm
      y: absolute y position in 1/10 mm
      z: absolute z position in 1/10 mm
      r: absolute r position in 1/10 mm
      g: absolute g position in 1/10 mm
      speed: speed select, 0 - slow, 1 - fast
      tw: target window class, set with STW

    Raises:
      TecanError: if moving to the target position causes a collision
    """

    cur_x = EVOArm._pos_cache.setdefault(self.module, await self.report_x_param(0))
    for module, pos in EVOArm._pos_cache.items():
      if module == self.module:
        continue
      if cur_x < x and cur_x < pos < x:  # moving right
        raise TecanError("Invalid command (collision)", self.module, 2)
      if cur_x > x and cur_x > pos > x:  # moving left
        raise TecanError("Invalid command (collision)", self.module, 2)
      if abs(pos - x) < 1500:
        raise TecanError("Invalid command (collision)", self.module, 2)

    await self.backend.send_command(
      module=self.module,
      command="SAA",
      params=[v, x, y, z, r, g, speed, 0, tw],
    )

  async def action_move_vector_coordinate_position(self):
    """Starts coordinate movement, built by vector coordinate table."""

    await self.backend.send_command(module=self.module, command="AAC")

    EVOArm._pos_cache[self.module] = await self.report_x_param(0)

  async def position_absolute_g(self, g: int):
    """Position absolute for G-axis

    Args:
      g: absolute position in 1/10 mm
    """

    await self.backend.send_command(module=self.module, command="PAG", params=[g])

  async def set_gripper_params(self, speed: int, pwm: int, cur: Optional[int] = None):
    """Set gripper parameters

    Args:
      speed: gripper search speed in 1/10 mm/s
      pwm: pulse width modification limit
      cur: maximal allowed current
    """

    await self.backend.send_command(module=self.module, command="SGG", params=[speed, pwm, cur])

  async def grip_plate(self, pos: int):
    """Grips plate at current X/Y/Z/R-position

    Args:
      pos: target position, plate must be fetched within current and target position
    """

    await self.backend.send_command(module=self.module, command="AGR", params=[pos])

  async def set_target_window_class(self, wc: int, x: int, y: int, z: int, r: int, g: int):
    """Sets drive parameters for the AAC command.

    Args:
      wc: window class, must be between 1 and 100
      x: target window for x-axis in 1/10 mm
      y: target window for y-axis in 1/10 mm
      z: target window for z-axis in 1/10 mm
      r: target window for r-axis in 1/10 deg
      g: target window for g-axis in 1/10 mm
    """

    await self.backend.send_command(module=self.module, command="STW", params=[wc, x, y, z, r, g])


# Deprecated alias with warning # TODO: remove mid May 2025 (giving people 1 month to update)
# https://github.com/PyLabRobot/pylabrobot/issues/466


class EVO(EVOBackend):
  def __init__(self, *args, **kwargs):
    raise RuntimeError(
      "`EVO` is deprecated. Please use `EVOBackend` instead.",
    )



import asyncio
import warnings

from pylabrobot.heating_shaking.backend import HeaterShakerBackend
from pylabrobot.io.serial import Serial
from pylabrobot.machines.backend import MachineBackend

try:
  import serial

  HAS_SERIAL = True
except ImportError as e:
  HAS_SERIAL = False
  _SERIAL_IMPORT_ERROR = e


class BioShake(HeaterShakerBackend):
  def __init__(self, port: str, timeout: int = 60):
    if not HAS_SERIAL:
      raise RuntimeError(
        f"pyserial is required for the BioShake module backend. Import error: {_SERIAL_IMPORT_ERROR}"
      )

    self.setup_finished = False
    self.port = port
    self.timeout = timeout
    self.io = Serial(
      port=self.port,
      baudrate=9600,
      bytesize=serial.EIGHTBITS,
      parity=serial.PARITY_NONE,
      stopbits=serial.STOPBITS_ONE,
      write_timeout=10,
      timeout=self.timeout,
      human_readable_device_name="BioShake",
    )

  async def _send_command(self, cmd: str, delay: float = 0.5, timeout: float = 2):
    try:
      # Flush serial buffers for a clean start
      await self.io.reset_input_buffer()
      await self.io.reset_output_buffer()

      # Send the command
      await self.io.write((cmd + "\r").encode("ascii"))
      await asyncio.sleep(delay)

      # Read and decode the response with a timeout
      try:
        response = await asyncio.wait_for(self.io.readline(), timeout=timeout)

      except asyncio.TimeoutError:
        raise RuntimeError(f"Timed out waiting for response to '{cmd}'")

      decoded = response.decode("ascii", errors="ignore").strip()

      # Parsing the response from the BioShake

      # No response at all
      if not decoded:
        raise RuntimeError(f"No response for '{cmd}'")

      # Device-specific errors
      if decoded.startswith("e"):
        raise RuntimeError(f"Device returned error for '{cmd}': '{decoded}'")

      if decoded.startswith("u ->"):
        raise NotImplementedError(f"'{cmd}' not supported: '{decoded}'")

      # Standard OK
      if decoded.lower().startswith("ok"):
        return None

      # All other valid responses (e.g. temperature and remaining time)
      return decoded

    except Exception as e:
      raise RuntimeError(f"Unexpected error while sending '{cmd}': {type(e).__name__}: {e}") from e

  async def setup(self, skip_home: bool = False):
    await MachineBackend.setup(self)
    await self.io.setup()
    if not skip_home:
      # Reset first before homing it to ensure the device is ready for run
      await self.reset()
      # Additional seconds until next command can be send after reset
      await asyncio.sleep(4)
      # Now home the device
      await self.home()

  async def stop(self):
    await MachineBackend.stop(self)
    await self.io.stop()

  async def reset(self):
    # Reset the BioShake if stuck in "e" state
    # Flush serial buffers for a clean start
    await self.io.reset_input_buffer()
    await self.io.reset_output_buffer()

    # Send the command
    await self.io.write(("resetDevice\r").encode("ascii"))

    start = asyncio.get_event_loop().time()
    max_seconds = 30  # How long a reset typically last

    while True:
      # Break the loop if process takes longer than 30 seconds
      if asyncio.get_event_loop().time() - start > max_seconds:
        raise TimeoutError("Reset did not complete in time")

      try:
        # Wait for each line with a timeout
        response = await asyncio.wait_for(self.io.readline(), timeout=2)
        decoded = response.decode("ascii", errors="ignore").strip()
        await asyncio.sleep(0.1)

        if len(decoded) > 0:
          # Stop when the final message arrives
          if "Initialization complete" in decoded:
            break

      except asyncio.TimeoutError:
        # Keep polling if nothing arrives within timeout
        continue

  async def home(self):
    # Initialize the BioShake into home position
    await self._send_command(cmd="shakeGoHome", delay=5)

  async def start_shaking(self, speed: float, acceleration: int = 0):
    # Check if speed is an integer
    if isinstance(speed, float):
      if not speed.is_integer():
        raise ValueError(f"Speed must be a whole number, not {speed}")
      speed = int(speed)
    if not isinstance(speed, int):
      raise TypeError(
        f"Speed must be an integer or a whole number float, not {type(speed).__name__}"
      )

    # Get the min and max speed of the device to assert speed
    min_speed = int(float(await self._send_command(cmd="getShakeMinRpm", delay=0.2)))
    max_speed = int(float(await self._send_command(cmd="getShakeMaxRpm", delay=0.2)))

    assert min_speed <= speed <= max_speed, (
      f"Speed {speed} RPM is out of range. Allowed range is {min_speed}{max_speed} RPM"
    )

    # Set the speed of the shaker
    set_speed_cmd = f"setShakeTargetSpeed{speed}"
    await self._send_command(cmd=set_speed_cmd)

    # Check if accel is an integer
    if isinstance(acceleration, float):
      if not acceleration.is_integer():  # type: ignore[attr-defined] # mypy is retarded
        raise ValueError(f"Acceleration must be a whole number, not {acceleration}")
      acceleration = int(acceleration)
    if not isinstance(acceleration, int):
      raise TypeError(
        f"Acceleration must be an integer or a whole number float, not {type(acceleration).__name__}"
      )

    # Get the min and max acceleration of the device to check bounds
    min_accel = int(float(await self._send_command(cmd="getShakeAccelerationMin", delay=0.2)))
    max_accel = int(float(await self._send_command(cmd="getShakeAccelerationMax", delay=0.2)))

    assert min_accel <= acceleration <= max_accel, (
      f"Acceleration {acceleration} seconds is out of range. Allowed range is {min_accel}-{max_accel} seconds"
    )

    # Set the acceleration of the shaker
    set_accel_cmd = f"setShakeAcceleration{acceleration}"
    await self._send_command(cmd=set_accel_cmd, delay=0.2)

    # Send the command to start shaking, either with or without duration

    await self._send_command(cmd="shakeOn", delay=0.2)

  async def shake(self, speed: float, acceleration: int = 0):
    warnings.warn(
      "BioShake.shake() is deprecated and will be removed in a future release. "
      "Use start_shaking() instead.",
      DeprecationWarning,
      stacklevel=2,
    )
    await self.start_shaking(speed=speed, acceleration=acceleration)

  async def stop_shaking(self, deceleration: int = 0):
    # Check if decel is an integer
    if isinstance(deceleration, float):
      if not deceleration.is_integer():  # type: ignore[attr-defined] # mypy is retarded
        raise ValueError(f"Deceleration must be a whole number, not {deceleration}")
      deceleration = int(deceleration)
    if not isinstance(deceleration, int):
      raise TypeError(
        f"Deceleration must be an integer or a whole number float, not {type(deceleration).__name__}"
      )

    # Get the min and max decel of the device to asset decel
    min_decel = int(float(await self._send_command(cmd="getShakeAccelerationMin", delay=0.2)))
    max_decel = int(float(await self._send_command(cmd="getShakeAccelerationMax", delay=0.2)))

    assert min_decel <= deceleration <= max_decel, (
      f"Deceleration {deceleration} seconds is out of range. Allowed range is {min_decel}-{max_decel} seconds"
    )

    # Set the deceleration of the shaker
    set_decel_cmd = f"setShakeAcceleration{deceleration}"
    await self._send_command(cmd=set_decel_cmd, delay=0.2)

    # stop shaking
    await self._send_command(cmd="shakeOff", delay=0.2)

    # The BioShake 3000 ELM firmware needs the motor to fully decelerate
    # before the edge-locking mechanism (ELM) can operate. Without this
    # delay, subsequent setElmUnlockPos commands return 'e' (error).
    sleep_time_after_stop = 3
    await asyncio.sleep(sleep_time_after_stop)

  @property
  def supports_locking(self) -> bool:
    return True

  async def lock_plate(self):
    await self._send_command(cmd="setElmLockPos", delay=0.3)

  async def unlock_plate(self):
    await self._send_command(cmd="setElmUnlockPos", delay=0.3)

  @property
  def supports_active_cooling(self) -> bool:
    return True

  async def set_temperature(self, temperature: float):
    # Get the min and max set points of the device to assert temperature
    min_temp = int(float(await self._send_command(cmd="getTempMin", delay=0.2)))
    max_temp = int(float(await self._send_command(cmd="getTempMax", delay=0.2)))

    assert min_temp <= temperature <= max_temp, (
      f"Temperature {temperature} C is out of range. Allowed range is {min_temp}–{max_temp} C."
    )

    temperature = temperature * 10

    # Check if temperature is an integer
    if isinstance(temperature, float):
      if not temperature.is_integer():
        raise ValueError(f"Temperature must be a whole number, not {temperature} (1/10 C)")
      temperature = int(temperature)
    if not isinstance(temperature, int):
      raise TypeError(
        f"Temperature must be an integer or a whole number float, not {type(temperature).__name__} (1/10 C)"
      )

    set_temp_cmd = f"setTempTarget{temperature}"
    await self._send_command(cmd=set_temp_cmd, delay=0.2)

    # Start temperature control
    await self._send_command(cmd="tempOn", delay=0.2)

  async def get_current_temperature(self) -> float:
    response = await self._send_command(cmd="getTempActual", delay=0.2)
    return float(response)

  async def deactivate(self):
    # Stop temperature control
    await self._send_command(cmd="tempOff", delay=0.2)



"""Tecan plates

size_z are best approximations, see
https://labautomation.io/t/pylabrobot-tecan-error-in-adding-labware-to-carrier/2987
"""

from typing import Dict, Optional

from pylabrobot.resources.plate import Lid, Plate
from pylabrobot.resources.tecan.tecan_resource import TecanResource
from pylabrobot.resources.utils import create_ordered_items_2d
from pylabrobot.resources.well import Well


class TecanPlate(Plate, TecanResource):
  """Base class for Tecan plates."""

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    z_start: float,
    z_dispense: float,
    z_max: float,
    area: float,
    ordered_items: Optional[Dict[str, Well]] = None,
    category: str = "tecan_plate",
    lid: Optional[Lid] = None,
    model: Optional[str] = None,
  ):
    super().__init__(
      name,
      size_x,
      size_y,
      size_z,
      ordered_items=ordered_items,
      category=category,
      lid=lid,
      model=model,
    )

    self.z_start = z_start
    self.z_dispense = z_dispense
    self.z_max = z_max
    self.area = area


def Microplate_96_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=129.9,
  #   size_y=83.9,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="Microplate_96_Well_Lid",
  # )


def Microplate_96_Well(name: str, with_lid: bool = False) -> TecanPlate:
  """white: pn 30122300, black: pn 30122298, cell culture/clear: pn 30122304, cell culture/black with clear bottom: pn 30122306

  Coley

  .. code-block:: python

      return TecanPlate(
        name=name,
        size_x=129.9,
        size_y=83.9,
        size_z=5.6,
        with_lid=with_lid,
        lid_height=8,
        model="Microplate_96_Well",
        z_start=1800.0,
        z_dispense=1970.0,
        z_max=2026.0,
        area=33.2,
        ordered_items=create_ordered_items_2d(Well,
          num_items_x=12,
          num_items_y=8,
          dx=10.8,
          dy=6.2,
          dz=0.0,
          item_dx=9.0,
          item_dy=9.0,
          size_x=9.0,
          size_y=9.0,
          size_z=22.6,
        ),
      )
  """

  return TecanPlate(
    name=name,
    size_x=127.8,
    size_y=85.4,
    size_z=7.6,
    lid=Microplate_96_Well_Lid(name=name + "_lid") if with_lid else None,
    model="Microplate_96_Well",
    z_start=1957.0,
    z_dispense=1975.0,
    z_max=2005.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=12,
      num_items_y=8,
      dx=12.5,
      dy=7.6,
      dz=0.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      size_z=22.6,
    ),
  )


def Microplate_portrait_96_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=85.4,
  #   size_y=127.8,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="Microplate_portrait_96_Well_Lid",
  # )


def Microplate_portrait_96_Well(name: str, with_lid: bool = False) -> TecanPlate:
  return TecanPlate(
    name=name,
    size_x=85.4,
    size_y=127.8,
    size_z=11.0,
    lid=Microplate_portrait_96_Well_Lid(name=name + "_lid") if with_lid else None,
    model="Microplate_portrait_96_Well",
    z_start=1940.0,
    z_dispense=1960.0,
    z_max=2050.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=8,
      num_items_y=12,
      dx=6.7,
      dy=9.9,
      dz=0.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      size_z=11.0,
    ),
  )


def DeepWell_96_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=127.8,
  #   size_y=85.4,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="DeepWell_96_Well_Lid",
  # )


def DeepWell_96_Well(name: str, with_lid: bool = False) -> TecanPlate:
  return TecanPlate(
    name=name,
    size_x=127.8,
    size_y=85.4,
    size_z=39.0,
    lid=DeepWell_96_Well_Lid(name=name + "_lid") if with_lid else None,
    model="DeepWell_96_Well",
    z_start=1670.0,
    z_dispense=1690.0,
    z_max=2060.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=12,
      num_items_y=8,
      dx=9.9,
      dy=6.7,
      dz=0.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      size_z=39.0,
    ),
  )


def HalfDeepWell_384_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=127.7,
  #   size_y=85.5,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="HalfDeepWell_384_Well_Lid",
  # )


def HalfDeepWell_384_Well(name: str, with_lid: bool = False) -> TecanPlate:
  return TecanPlate(
    name=name,
    size_x=127.7,
    size_y=85.5,
    size_z=18.8,
    lid=HalfDeepWell_384_Well_Lid(name=name + "_lid") if with_lid else None,
    model="HalfDeepWell_384_Well",
    z_start=1869.0,
    z_dispense=1889.0,
    z_max=2057.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=24,
      num_items_y=16,
      dx=9.85,
      dy=6.75,
      dz=0.0,
      item_dx=4.5,
      item_dy=4.5,
      size_x=4.5,
      size_y=4.5,
      size_z=18.8,
    ),
  )


def DeepWell_portrait_96_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=85.4,
  #   size_y=127.8,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="DeepWell_portrait_96_Well_Lid",
  # )


def DeepWell_portrait_96_Well(name: str, with_lid: bool = False) -> TecanPlate:
  return TecanPlate(
    name=name,
    size_x=85.4,
    size_y=127.8,
    size_z=38.0,
    lid=DeepWell_portrait_96_Well_Lid(name=name + "_lid") if with_lid else None,
    model="DeepWell_portrait_96_Well",
    z_start=1670.0,
    z_dispense=1690.0,
    z_max=2050.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=8,
      num_items_y=12,
      dx=6.7,
      dy=9.9,
      dz=0.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      size_z=38.0,
    ),
  )


def Plate_portrait_384_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=85.5,
  #   size_y=127.7,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="Plate_portrait_384_Well_Lid",
  # )


def Plate_portrait_384_Well(name: str, with_lid: bool = False) -> TecanPlate:
  return TecanPlate(
    name=name,
    size_x=85.5,
    size_y=127.7,
    size_z=11.0,
    lid=Plate_portrait_384_Well_Lid(name=name + "_lid") if with_lid else None,
    model="Plate_portrait_384_Well",
    z_start=1940.0,
    z_dispense=1960.0,
    z_max=2050.0,
    area=9.0,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=16,
      num_items_y=24,
      dx=6.75,
      dy=9.85,
      dz=0.0,
      item_dx=4.5,
      item_dy=4.5,
      size_x=4.5,
      size_y=4.5,
      size_z=11.0,
    ),
  )


def Macherey_Nagel_Plate_96_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=151.6,
  #   size_y=131.1,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="Macherey_Nagel_Plate_96_Well_Lid",
  # )


def Macherey_Nagel_Plate_96_Well(name: str, with_lid: bool = False) -> TecanPlate:
  return TecanPlate(
    name=name,
    size_x=151.6,
    size_y=131.1,
    size_z=29.9,
    lid=Macherey_Nagel_Plate_96_Well_Lid(name=name + "_lid") if with_lid else None,
    model="Macherey_Nagel_Plate_96_Well",
    z_start=1532.0,
    z_dispense=1578.0,
    z_max=1831.0,
    area=65.0,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=12,
      num_items_y=8,
      dx=22.25,
      dy=29.75,
      dz=0.0,
      item_dx=8.9,
      item_dy=8.9,
      size_x=8.9,
      size_y=8.9,
      size_z=29.9,
    ),
  )


def Qiagen_Plate_96_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=151.7,
  #   size_y=132.0,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="Qiagen_Plate_96_Well_Lid",
  # )


def Qiagen_Plate_96_Well(name: str, with_lid: bool = False) -> TecanPlate:
  return TecanPlate(
    name=name,
    size_x=151.7,
    size_y=132.0,
    size_z=26.6,
    lid=Qiagen_Plate_96_Well_Lid(name=name + "_lid") if with_lid else None,
    model="Qiagen_Plate_96_Well",
    z_start=1541.0,
    z_dispense=1549.0,
    z_max=1807.0,
    area=60.8,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=12,
      num_items_y=8,
      dx=22.25,
      dy=30.05,
      dz=0.0,
      item_dx=8.9,
      item_dy=8.9,
      size_x=8.9,
      size_y=8.9,
      size_z=26.6,
    ),
  )


def AB_Plate_96_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=130.9,
  #   size_y=128.8,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="AB_Plate_96_Well_Lid",
  # )


def AB_Plate_96_Well(name: str, with_lid: bool = False) -> TecanPlate:
  return TecanPlate(
    name=name,
    size_x=130.9,
    size_y=128.8,
    size_z=19.5,
    lid=AB_Plate_96_Well_Lid(name=name + "_lid") if with_lid else None,
    model="AB_Plate_96_Well",
    z_start=1822.0,
    z_dispense=1837.0,
    z_max=2017.0,
    area=26.4,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=12,
      num_items_y=8,
      dx=11.4,
      dy=28.0,
      dz=-0.5,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      size_z=19.5,
    ),
  )


def PCR_Plate_96_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=128.0,
  #   size_y=83.2,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="PCR_Plate_96_Well_Lid",
  # )


def PCR_Plate_96_Well(name: str, with_lid: bool = False) -> TecanPlate:
  return TecanPlate(
    name=name,
    size_x=128.0,
    size_y=83.2,
    size_z=19.5,
    lid=PCR_Plate_96_Well_Lid(name=name + "_lid") if with_lid else None,
    model="PCR_Plate_96_Well",
    z_start=1900.0,
    z_dispense=1915.0,
    z_max=2095.0,
    area=28.3,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=12,
      num_items_y=8,
      dx=10.1,
      dy=5.3,
      dz=0.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      size_z=19.5,
    ),
  )


def DeepWell_Greiner_1536_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=127.8,
  #   size_y=85.5,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="DeepWell_Greiner_1536_Well_Lid",
  # )


def DeepWell_Greiner_1536_Well(name: str, with_lid: bool = False) -> TecanPlate:
  return TecanPlate(
    name=name,
    size_x=127.8,
    size_y=85.5,
    size_z=8.6,
    lid=DeepWell_Greiner_1536_Well_Lid(name=name + "_lid") if with_lid else None,
    model="DeepWell_Greiner_1536_Well",
    z_start=1984.0,
    z_dispense=2004.0,
    z_max=2070.0,
    area=2.7,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=48,
      num_items_y=32,
      dx=9.85,
      dy=6.75,
      dz=0.0,
      item_dx=2.3,
      item_dy=2.3,
      size_x=2.3,
      size_y=2.3,
      size_z=8.6,
    ),
  )


def Hibase_Greiner_1536_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=127.8,
  #   size_y=85.5,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="Hibase_Greiner_1536_Well_Lid",
  # )


def Hibase_Greiner_1536_Well(name: str, with_lid: bool = False) -> TecanPlate:
  return TecanPlate(
    name=name,
    size_x=127.8,
    size_y=85.5,
    size_z=5.4,
    lid=Hibase_Greiner_1536_Well_Lid(name=name + "_lid") if with_lid else None,
    model="Hibase_Greiner_1536_Well",
    z_start=1984.0,
    z_dispense=2004.0,
    z_max=2038.0,
    area=2.5,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=48,
      num_items_y=32,
      dx=9.85,
      dy=6.75,
      dz=0.0,
      item_dx=2.3,
      item_dy=2.3,
      size_x=2.3,
      size_y=2.3,
      size_z=5.4,
    ),
  )


def Lowbase_Greiner_1536_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=127.8,
  #   size_y=85.5,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="Lowbase_Greiner_1536_Well_Lid",
  # )


def Lowbase_Greiner_1536_Well(name: str, with_lid: bool = False) -> TecanPlate:
  return TecanPlate(
    name=name,
    size_x=127.8,
    size_y=85.5,
    size_z=6.2,
    lid=Lowbase_Greiner_1536_Well_Lid(name=name + "_lid") if with_lid else None,
    model="Lowbase_Greiner_1536_Well",
    z_start=2024.0,
    z_dispense=2034.0,
    z_max=2086.0,
    area=2.7,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=48,
      num_items_y=32,
      dx=9.85,
      dy=6.75,
      dz=0.0,
      item_dx=2.3,
      item_dy=2.3,
      size_x=2.3,
      size_y=2.3,
      size_z=6.2,
    ),
  )


def Separation_Plate_96_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=151.7,
  #   size_y=132.0,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="Separation_Plate_96_Well_Lid",
  # )


def Separation_Plate_96_Well(name: str, with_lid: bool = False) -> TecanPlate:
  return TecanPlate(
    name=name,
    size_x=151.7,
    size_y=132.0,
    size_z=26.6,
    lid=Separation_Plate_96_Well_Lid(name=name + "_lid") if with_lid else None,
    model="Separation_Plate_96_Well",
    z_start=1541.0,
    z_dispense=1549.0,
    z_max=1807.0,
    area=60.8,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=12,
      num_items_y=8,
      dx=22.25,
      dy=30.05,
      dz=0.0,
      item_dx=8.9,
      item_dy=8.9,
      size_x=8.9,
      size_y=8.9,
      size_z=26.6,
    ),
  )


def DeepWell_square_96_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=127.8,
  #   size_y=85.4,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="DeepWell_square_96_Well_Lid",
  # )


def DeepWell_square_96_Well(name: str, with_lid: bool = False) -> TecanPlate:
  return TecanPlate(
    name=name,
    size_x=127.8,
    size_y=85.4,
    size_z=39.0,
    lid=DeepWell_square_96_Well_Lid(name=name + "_lid") if with_lid else None,
    model="DeepWell_square_96_Well",
    z_start=1670.0,
    z_dispense=1690.0,
    z_max=2060.0,
    area=64.0,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=12,
      num_items_y=8,
      dx=9.9,
      dy=6.7,
      dz=0.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      size_z=39.0,
    ),
  )


def CaCo2_Plate_24_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=125.2,
  #   size_y=89.2,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="CaCo2_Plate_24_Well_Lid",
  # )


def CaCo2_Plate_24_Well(name: str, with_lid: bool = False) -> TecanPlate:
  return TecanPlate(
    name=name,
    size_x=125.2,
    size_y=89.2,
    size_z=6.5,
    lid=CaCo2_Plate_24_Well_Lid(name=name + "_lid") if with_lid else None,
    model="CaCo2_Plate_24_Well",
    z_start=1960.0,
    z_dispense=2007.0,
    z_max=2025.0,
    area=50.3,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=6,
      num_items_y=4,
      dx=4.75,
      dy=5.95,
      dz=0.0,
      item_dx=19.3,
      item_dy=19.3,
      size_x=19.3,
      size_y=19.3,
      size_z=6.5,
    ),
  )


def Plate_384_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=127.7,
  #   size_y=85.5,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="Plate_384_Well_Lid",
  # )


def Plate_384_Well(name: str, with_lid: bool = False) -> TecanPlate:
  """white: pn 30122301, black: pn 30122299, cell culture/clear: pn 30122305, cell culture/black with clear bottom: pn 30122307"""
  return TecanPlate(
    name=name,
    size_x=127.7,
    size_y=85.5,
    size_z=12.1,
    lid=Plate_384_Well_Lid(name=name + "_lid") if with_lid else None,
    model="Plate_384_Well",
    z_start=1940.0,
    z_dispense=1960.0,
    z_max=2061.0,
    area=13.7,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=24,
      num_items_y=16,
      dx=9.85,
      dy=6.75,
      dz=0.0,
      item_dx=4.5,
      item_dy=4.5,
      size_x=4.5,
      size_y=4.5,
      size_z=12.1,
    ),
  )


def Microplate_24_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=130.9,
  #   size_y=85.5,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="Microplate_24_Well_Lid",
  # )


def Microplate_24_Well(name: str, with_lid: bool = False) -> TecanPlate:
  """cell culture/clear: pn 30122302"""
  return TecanPlate(
    name=name,
    size_x=130.9,
    size_y=85.5,
    size_z=17.6,
    lid=Microplate_24_Well_Lid(name=name + "_lid") if with_lid else None,
    model="Microplate_24_Well",
    z_start=1885.0,
    z_dispense=1917.0,
    z_max=2061.0,
    area=193.6,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=6,
      num_items_y=4,
      dx=6.6,
      dy=3.5,
      dz=0.0,
      item_dx=19.6,
      item_dy=19.6,
      size_x=19.6,
      size_y=19.6,
      size_z=17.6,
    ),
  )


def TecanExtractionPlate_96_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=129.8,
  #   size_y=91.7,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="TecanExtractionPlate_96_Well_Lid",
  # )


def TecanExtractionPlate_96_Well(name: str, with_lid: bool = False) -> TecanPlate:
  return TecanPlate(
    name=name,
    size_x=129.8,
    size_y=91.7,
    size_z=23.0,
    lid=TecanExtractionPlate_96_Well_Lid(name=name + "_lid") if with_lid else None,
    model="TecanExtractionPlate_96_Well",
    z_start=1831.0,
    z_dispense=1910.0,
    z_max=2061.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=12,
      num_items_y=8,
      dx=10.9,
      dy=9.7,
      dz=0.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      size_z=23.0,
    ),
  )


def Microplate_48_Well_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")
  # See https://github.com/PyLabRobot/pylabrobot/pull/161.
  # return Lid(
  #   name=name,
  #   size_x=131.1,
  #   size_y=85.3,
  #   size_z=None,           # measure the total z height
  #   nesting_z_height=None, # measure overlap between lid and plate
  #   model="Microplate_48_Well_Lid",
  # )


def Microplate_48_Well(name: str, with_lid: bool = False) -> TecanPlate:
  """cell culture/clear: pn 30122303"""
  return TecanPlate(
    name=name,
    size_x=131.1,
    size_y=85.3,
    size_z=17.3,
    lid=Microplate_48_Well_Lid(name=name + "_lid") if with_lid else None,
    model="Microplate_48_Well",
    z_start=1887.0,
    z_dispense=1921.0,
    z_max=2060.0,
    area=102.1,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=8,
      num_items_y=6,
      dx=13.5,
      dy=3.8,
      dz=0.0,
      item_dx=13.0,
      item_dy=13.0,
      size_x=13.0,
      size_y=13.0,
      size_z=17.3,
    ),
  )



import asyncio
import unittest
from unittest.mock import AsyncMock, MagicMock

from pylabrobot.resources import Coordinate
from pylabrobot.thermocycling import (
  Thermocycler,
  ThermocyclerBackend,
  ThermocyclerChatterboxBackend,
)
from pylabrobot.thermocycling.standard import Protocol, Stage, Step


def mock_backend() -> MagicMock:
  """Creates a fully compliant mock of the ThermocyclerBackend using a spec."""
  mock = MagicMock(spec=ThermocyclerBackend)
  mock.setup = AsyncMock()
  mock.stop = AsyncMock()
  mock.open_lid = AsyncMock()
  mock.close_lid = AsyncMock()
  mock.set_block_temperature = AsyncMock()
  mock.set_lid_temperature = AsyncMock()
  mock.deactivate_block = AsyncMock()
  mock.deactivate_lid = AsyncMock()
  mock.run_protocol = AsyncMock()
  mock.get_block_current_temperature = AsyncMock(return_value=[25.0])
  mock.get_block_target_temperature = AsyncMock(return_value=None)
  mock.get_lid_current_temperature = AsyncMock(return_value=[25.0])
  mock.get_lid_target_temperature = AsyncMock(return_value=None)
  mock.get_lid_open = AsyncMock(return_value=False)
  mock.get_lid_temperature_status = AsyncMock(return_value="idle")
  mock.get_block_status = AsyncMock(return_value="idle")
  mock.get_hold_time = AsyncMock(return_value=0.0)
  mock.get_current_cycle_index = AsyncMock(return_value=0)
  mock.get_total_cycle_count = AsyncMock(return_value=0)
  mock.get_current_step_index = AsyncMock(return_value=0)
  mock.get_total_step_count = AsyncMock(return_value=0)
  return mock


class ThermocyclerTests(unittest.IsolatedAsyncioTestCase):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.tc = Thermocycler(
      name="test_tc",
      size_x=10,
      size_y=10,
      size_z=10,
      backend=mock_backend(),
      child_location=Coordinate(0, 0, 0),
    )

  def test_thermocycler_serialization(self):
    """Test that the high-level resource serializes and deserializes correctly."""
    self.tc.backend = ThermocyclerChatterboxBackend()
    serialized = self.tc.serialize()
    deserialized = Thermocycler.deserialize(serialized)
    assert self.tc == deserialized

  async def test_run_pcr_profile_builds_correct_profile(self):
    """Test that run_pcr_profile correctly builds the flat step list."""

    async def mock_wait_for_lid(*args, **kwargs):
      pass

    self.tc.wait_for_lid = mock_wait_for_lid  # type: ignore

    await self.tc.run_pcr_profile(
      denaturation_temp=[98.0],
      denaturation_time=10.0,
      annealing_temp=[55.0],
      annealing_time=30.0,
      extension_temp=[72.0],
      extension_time=60.0,
      num_cycles=2,
      block_max_volume=25.0,
      lid_temperature=[105.0],
      pre_denaturation_temp=[95.0],
      pre_denaturation_time=180.0,
      final_extension_temp=[72.0],
      final_extension_time=300.0,
      storage_temp=[4.0],
      storage_time=600.0,
    )

    self.tc.backend.set_lid_temperature.assert_called_once_with([105.0])  # type: ignore

    expected_protocol = Protocol(
      stages=[
        Stage(steps=[Step(temperature=[95.0], hold_seconds=180.0, rate=None)], repeats=1),
        Stage(
          steps=[
            Step(temperature=[98.0], hold_seconds=10.0, rate=None),
            Step(temperature=[55.0], hold_seconds=30.0, rate=None),
            Step(temperature=[72.0], hold_seconds=60.0, rate=None),
          ],
          repeats=2,
        ),
        Stage(steps=[Step(temperature=[72.0], hold_seconds=300.0, rate=None)], repeats=1),
        Stage(steps=[Step(temperature=[4.0], hold_seconds=600.0, rate=None)], repeats=1),
      ]
    )

    self.tc.backend.run_protocol.assert_called_once_with(expected_protocol, 25.0)  # type: ignore

  async def test_wait_for_profile_completion(self):
    """Test that wait_for_profile_completion correctly polls is_profile_running."""
    self.tc.backend.get_hold_time.side_effect = [10.0, 5.0, 0.0]  # type: ignore

    # Patch asyncio.sleep to a no-op for the test.
    original_sleep = asyncio.sleep

    async def mock_sleep(*args, **kwargs):
      pass

    asyncio.sleep = mock_sleep
    try:
      await self.tc.wait_for_profile_completion(poll_interval=0.01)
      assert self.tc.backend.get_hold_time.call_count == 3  # type: ignore
    finally:
      asyncio.sleep = original_sleep

  async def test_is_profile_running_logic(self):
    """Test that `is_profile_running` returns the correct boolean based on various profile states."""
    test_cases = [
      (10.0, 1, 10, 1, 3, True),
      (0.0, 5, 10, 1, 3, True),
      (0.0, 9, 10, 1, 3, True),
      (0.0, 9, 10, 3, 3, False),
      (0.0, 1, 1, 1, 1, False),
    ]
    for hold, cycle, total_cycles, step, total_steps, expected in test_cases:
      self.tc.backend.get_hold_time.return_value = hold  # type: ignore
      self.tc.backend.get_current_cycle_index.return_value = cycle  # type: ignore
      self.tc.backend.get_total_cycle_count.return_value = total_cycles  # type: ignore
      self.tc.backend.get_current_step_index.return_value = step  # type: ignore
      self.tc.backend.get_total_step_count.return_value = total_steps  # type: ignore
      print(f"Testing with hold={hold}, cycle={cycle}, total_cycles={total_cycles}, ")
      assert await self.tc.is_profile_running() is expected



from __future__ import annotations

import warnings
from collections import OrderedDict
from typing import (
  TYPE_CHECKING,
  Dict,
  List,
  Literal,
  Optional,
  Sequence,
  Tuple,
  Union,
  cast,
)

from pylabrobot.resources.liquid import Liquid
from pylabrobot.resources.resource_holder import get_child_location

from .itemized_resource import ItemizedResource
from .resource import Coordinate, Resource

if TYPE_CHECKING:
  from .well import Well


class Lid(Resource):
  """Lid for plates."""

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    nesting_z_height: float,
    category: str = "lid",
    model: Optional[str] = None,
  ):
    """Create a lid for a plate.

    Args:
      name: Name of the lid.
      size_x: Size of the lid in x-direction.
      size_y: Size of the lid in y-direction.
      size_z: Size of the lid in z-direction.
      nesting_z_height: the overlap in mm between the lid and its parent plate (in the z-direction).
    """
    super().__init__(
      name=name,
      size_x=size_x,
      size_y=size_y,
      size_z=size_z,
      category=category,
      model=model,
    )
    self.nesting_z_height = nesting_z_height
    if nesting_z_height == 0:
      print(f"{self.name}: Are you certain that the lid nests 0 mm with its parent plate?")

  def serialize(self) -> dict:
    return {
      **super().serialize(),
      "nesting_z_height": self.nesting_z_height,
    }


class Plate(ItemizedResource["Well"]):
  """Base class for Plate resources."""

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    ordered_items: Optional[Dict[str, "Well"]] = None,
    ordering: Optional[OrderedDict[str, str]] = None,
    category: str = "plate",
    lid: Optional[Lid] = None,
    model: Optional[str] = None,
    plate_type: Literal["skirted", "semi-skirted", "non-skirted"] = "skirted",
  ):
    """Initialize a Plate resource.

    Args:
      well_size_x: Size of the wells in the x direction.
      well_size_y: Size of the wells in the y direction.
      lid: Immediately assign a lid to the plate.
      plate_type: Type of the plate. One of "skirted", "semi-skirted", or "non-skirted". A
        WIP: https://github.com/PyLabRobot/pylabrobot/pull/152#discussion_r1625831517
    """

    super().__init__(
      name,
      size_x,
      size_y,
      size_z,
      ordered_items=ordered_items,
      ordering=ordering,
      category=category,
      model=model,
    )
    self._lid: Optional[Lid] = None
    self.plate_type = plate_type

    if lid is not None:
      self.assign_child_resource(lid)

  @property
  def lid(self) -> Optional[Lid]:
    return self._lid

  @lid.setter
  def lid(self, lid: Optional[Lid]) -> None:
    if lid is None:
      self.unassign_child_resource(self._lid)
    else:
      self.assign_child_resource(lid)
    self._lid = lid

  def get_lid_location(self, lid: Lid) -> Coordinate:
    """Get location of the lid when assigned to the plate. Takes into account sinking and rotation."""
    return get_child_location(lid) + Coordinate(0, 0, self.get_size_z() - lid.nesting_z_height)

  def assign_child_resource(
    self,
    resource: Resource,
    location: Optional[Coordinate] = None,
    reassign: bool = True,
  ):
    if isinstance(resource, Lid):
      if self.has_lid():
        raise ValueError(f"Plate '{self.name}' already has a lid.")
      self._lid = resource
      default_location = self.get_lid_location(resource)
      location = location or default_location
    else:
      assert location is not None, "Location must be specified for if resource is not a lid."
    return super().assign_child_resource(resource, location=location, reassign=reassign)

  def unassign_child_resource(self, resource):
    if isinstance(resource, Lid) and resource == self.lid:
      self._lid = None
    return super().unassign_child_resource(resource)

  def __repr__(self) -> str:
    return (
      f"{self.__class__.__name__}(name={self.name!r}, size_x={self._size_x}, "
      f"size_y={self._size_y}, size_z={self._size_z}, location={self.location})"
    )

  def get_well(self, identifier: Union[str, int, Tuple[int, int]]) -> "Well":
    """Get the item with the given identifier.

    See :meth:`~.get_item` for more information.
    """

    return super().get_item(identifier)

  def get_wells(self, identifier: Union[str, Sequence[int]]) -> List["Well"]:
    """Get the wells with the given identifier.

    See :meth:`~.get_items` for more information.
    """

    return super().get_items(identifier)

  def has_lid(self) -> bool:
    return self.lid is not None

  def set_well_volumes(
    self,
    volumes: List[float],
  ) -> None:
    """Fill all wells in the plate with a given volume.

    Args:
      volumes: The volume to fill each well with, in uL.
    """

    if not len(volumes) == self.num_items:
      raise ValueError(
        f"Length of volumes ({len(volumes)}) does not match number of wells ({self.num_items})."
      )

    for well, volume in zip(self.get_all_items(), volumes):
      well.set_volume(volume)

  def set_well_liquids(
    self,
    liquids: Union[
      List[List[Tuple[Optional["Liquid"], Union[int, float]]]],
      List[Tuple[Optional["Liquid"], Union[int, float]]],
      Tuple[Optional["Liquid"], Union[int, float]],
    ],
  ):
    """Deprecated: Use `set_well_volumes` instead."""
    warnings.warn(
      "set_well_liquids is deprecated and will be removed in a future version. "
      "Use set_well_volumes instead.",
      FutureWarning,
    )
    if isinstance(liquids, tuple):
      liquids = [liquids] * self.num_items
    elif isinstance(liquids, list) and all(isinstance(column, list) for column in liquids):
      # mypy doesn't know that all() checks the type
      liquids = cast(List[List[Tuple[Optional["Liquid"], float]]], liquids)
      liquids = [list(column) for column in zip(*liquids)]  # transpose the list of lists
      liquids = [volume for column in liquids for volume in column]  # flatten the list of lists

    self.set_well_volumes([volume for _, volume in liquids])  # type: ignore

  def disable_volume_trackers(self) -> None:
    """Disable volume tracking for all wells in the plate."""

    for well in self.get_all_items():
      well.tracker.disable()

  def enable_volume_trackers(self) -> None:
    """Enable volume tracking for all wells in the plate."""

    for well in self.get_all_items():
      well.tracker.enable()

  def get_quadrant(
    self,
    quadrant: Literal[
      "tl", "top_left", "tr", "top_right", "bl", "bottom_left", "br", "bottom_right"
    ],
    quadrant_type: Literal["block", "checkerboard"] = "checkerboard",
    quadrant_internal_fill_order: Literal["column-major", "row-major"] = "column-major",
  ) -> List["Well"]:
    """
    Get wells from a specified quadrant.

    Args:
      quadrant: The desired quadrant ("tl" / "top_left", "tr" / "top_right",
        "bl" / "bottom_left", "br" / "bottom_right").
      quadrant_type: Either "block" (divides plate into 4 sections) or
        "checkerboard" (alternating well pattern).
      quadrant_internal_fill_order: Whether to return wells in "column-major"
        or "row-major" order.

    Returns:
      List of wells in the specified quadrant.

    Raises:
      ValueError: If an invalid quadrant or configuration is specified.
    """

    # Ensure plate dimensions are even for valid quadrant selection
    if self.num_items_x % 2 != 0 or self.num_items_y % 2 != 0:
      raise ValueError(
        "Both num_items_x and num_items_y must be even for quadrant selection,"
        f"\nare {self.num_items_x=}, {self.num_items_y=}"
      )
    assert quadrant_internal_fill_order in ["column-major", "row-major"], (
      f"Invalid quadrant_internal_fill_order: {quadrant_internal_fill_order},"
      "\nquadrant_internal_fill_order must be either 'column-major' or 'row-major',"
    )

    # Determine row and column start indices
    if quadrant.lower() in ["tl", "top_left"]:
      row_start, col_start = 0, 0
    elif quadrant.lower() in ["tr", "top_right"]:
      row_start, col_start = 0, 1
    elif quadrant.lower() in ["bl", "bottom_left"]:
      row_start, col_start = 1, 0
    elif quadrant.lower() in ["br", "bottom_right"]:
      row_start, col_start = 1, 1
    else:
      raise ValueError(
        f"Invalid quadrant: {quadrant}. Quadrant must be in  ['tl', 'tr', 'bl', 'br']"
      )

    wells = []

    if quadrant_type == "checkerboard":
      # Checkerboard pattern: Every other well
      for row in range(row_start, self.num_items_y, 2):
        for col in range(col_start, self.num_items_x, 2):
          wells.append(self.get_well((row, col)))

    elif quadrant_type == "block":
      # Block pattern: Ensure plate can be evenly divided into 4 quadrants
      row_half = self.num_items_y // 2
      col_half = self.num_items_x // 2

      row_range = range(row_start * row_half, (row_start + 1) * row_half)
      col_range = range(col_start * col_half, (col_start + 1) * col_half)

      for row in row_range:
        for col in col_range:
          wells.append(self.get_well((row, col)))

    else:
      raise ValueError(
        f"Invalid quadrant_type: {quadrant_type}. "
        "quadrant_type must be either 'checkerboard' (default) or 'block'"
      )

    # Apply internal fill order
    assert all(well.location is not None for well in wells)
    if quadrant_internal_fill_order == "row-major":
      wells.sort(key=lambda well: (-well.location.y, well.location.x))  # type: ignore
    else:
      wells.sort(key=lambda well: (well.location.x, -well.location.y))  # type: ignore

    return wells

  def check_can_drop_resource_here(self, resource: Resource, *, reassign: bool = True) -> None:
    if not isinstance(resource, Lid):
      raise RuntimeError(f"Can only drop Lid resources onto Plate '{self.name}'.")



"""Tests for Coordinate"""

import unittest

from pylabrobot.serializer import deserialize, serialize

from .coordinate import Coordinate


class TestCoordinate(unittest.TestCase):
  def setUp(self):
    self.a = Coordinate(1, 2, 3)
    self.b = Coordinate(10, 10, 10)
    self.c = Coordinate(0, 0, 0)

  def test_addition(self):
    self.assertEqual(self.a, self.a)
    self.assertEqual(self.a + self.c, self.a)
    self.assertEqual(self.a + self.b, Coordinate(11, 12, 13))
    self.assertEqual(self.b + self.b, Coordinate(20, 20, 20))

  def test_to_string(self):
    self.assertEqual(f"{self.a}", "Coordinate(001.000, 002.000, 003.000)")

  def test_serialization(self):
    self.assertEqual(
      serialize(self.a),
      {"x": 1, "y": 2, "z": 3, "type": "Coordinate"},
    )
    self.assertEqual(self.a, deserialize(serialize(self.a)))

  def test_unpacking(self):
    x, y, z = self.a
    self.assertEqual((x, y, z), (1, 2, 3))



import pytest

from pylabrobot.resources.coordinate import Coordinate
from pylabrobot.resources.resource import Resource
from pylabrobot.resources.utils import (
  label_to_row_index,
  query,
  row_index_to_label,
  split_identifier,
)
from pylabrobot.resources.well import Well


def test_query():
  root = Resource(name="root", size_x=10, size_y=10, size_z=10)
  child1 = Resource(name="child1", size_x=5, size_y=5, size_z=5)
  child2 = Resource(name="child2", size_x=5, size_y=5, size_z=5)
  child3 = Resource(name="odd", size_x=5, size_y=5, size_z=5)
  root.assign_child_resource(child1, location=Coordinate(0, 0, 0))
  root.assign_child_resource(child2, location=Coordinate(5, 0, 0))
  root.assign_child_resource(child3, location=Coordinate(0, 5, 0))

  assert query(root, Resource) == [child1, child2, child3]
  assert query(root, Resource, x=0) == [child1, child3]
  assert query(root, Resource, y=0) == [child1, child2]
  assert query(root, name=r"child\d") == [child1, child2]


def test_query_with_type():
  root = Resource(name="root", size_x=10, size_y=10, size_z=10)
  well1 = Well(name="well1", size_x=3, size_y=3, size_z=3)
  well2 = Well(name="well2", size_x=3, size_y=3, size_z=3)
  root.assign_child_resource(well1, location=Coordinate(6, 1, 0))
  root.assign_child_resource(well2, location=Coordinate(6, 6, 0))
  assert query(root, Well) == [well1, well2]
  assert query(root, Well, x=6) == [well1, well2]


class TestRowLabels:
  def test_single_letter(self):
    assert row_index_to_label(0) == "A"
    assert row_index_to_label(7) == "H"
    assert row_index_to_label(25) == "Z"

  def test_double_letter(self):
    assert row_index_to_label(26) == "AA"
    assert row_index_to_label(27) == "AB"
    assert row_index_to_label(31) == "AF"

  def test_negative_raises(self):
    with pytest.raises(ValueError):
      row_index_to_label(-1)

  def test_roundtrip(self):
    for i in range(52):
      assert label_to_row_index(row_index_to_label(i)) == i

  def test_label_to_index(self):
    assert label_to_row_index("A") == 0
    assert label_to_row_index("Z") == 25
    assert label_to_row_index("AA") == 26
    assert label_to_row_index("AF") == 31

  def test_label_case_insensitive(self):
    assert label_to_row_index("a") == 0
    assert label_to_row_index("af") == 31

  def test_split_identifier(self):
    assert split_identifier("A1") == ("A", "1")
    assert split_identifier("AF48") == ("AF", "48")
    assert split_identifier("P24") == ("P", "24")

  def test_split_identifier_no_digits_raises(self):
    with pytest.raises(ValueError):
      split_identifier("ABC")


def test_deep():
  root = Resource(name="root", size_x=10, size_y=10, size_z=10)
  child1 = Resource(name="child1", size_x=5, size_y=5, size_z=5)
  grandchild = Resource(name="grandchild", size_x=2, size_y=2, size_z=2)
  root.assign_child_resource(child1, location=Coordinate(0, 0, 0))
  child1.assign_child_resource(grandchild, location=Coordinate(1, 1, 0))

  assert query(root, Resource) == [child1, grandchild]



import unittest

from pylabrobot.heating_shaking import HeaterShaker, HeaterShakerChatterboxBackend
from pylabrobot.resources.coordinate import Coordinate


class HeaterShakerTests(unittest.TestCase):
  def test_serialization(self):
    hs = HeaterShaker(
      name="test_hs",
      size_x=10,
      size_y=10,
      size_z=10,
      backend=HeaterShakerChatterboxBackend(),
      child_location=Coordinate(0, 0, 0),
    )

    serialized = hs.serialize()
    deserialized = HeaterShaker.deserialize(serialized)
    self.assertEqual(hs, deserialized)



import logging
from typing import List, Optional

from pylabrobot.resources.coordinate import Coordinate
from pylabrobot.resources.plate import Plate
from pylabrobot.resources.resource import Resource
from pylabrobot.resources.resource_holder import get_child_location

logger = logging.getLogger("pylabrobot")


class ResourceStack(Resource):
  """ResourceStack represent a group of resources that are stacked together and act as a single
  unit. Stacks can grow be configured to be able to grow in x, y, or z direction. Stacks growing
  in the x direction are from left to right. Stacks growing in the y direction are from front to
  back. Stacks growing in the z direction are from bottom to top, and function as the
  `stack data type <https://en.wikipedia.org/wiki/Stack_(abstract_data_type)>`.

  Attributes:
    name: The name of the resource group.
    location: The location of the resource group. This will be the location of the first resource in
      the group.
    category: The category of the resource group.
    resources: The resources in the resource group.

  Examples:
    Making a resource group containing two resources:

    >>> stack = ResourceStack(“patched_plate”, "z", [
    ...   Resource("plate1", size_x=1, size_y=1, size_z=10),
    ...   Resource("plate2", size_x=1, size_y=1, size_z=10),
    ... ])
    >>> stack.get_size_x()
    1
    >>> stack.get_size_z()
    30

    :meth:`Moving <pyhamilton.liquid_handling.LiquidHandler.move_plate>` a plate to the a stacking
    area.

    >>> lh.move_plate(plate, stacking_area)

    :meth:`Moving <pyhamilton.liquid_handling.LiquidHandler.move_lid>` a lid to the stacking area.

    >>> lh.move_lid(plate.lid, stacking_area)

    Getting a plate from the stacking area and moving it to a :class:`~abstract.PlateCarrier`.

    >>> lh.move_plate(stacking_area.get_top_item(), plt_car[0])
  """

  def __init__(
    self,
    name: str,
    direction: str,
    resources: Optional[List[Resource]] = None,
  ):
    super().__init__(name, size_x=0, size_y=0, size_z=0, category="resource_group")
    assert direction in [
      "x",
      "y",
      "z",
    ], "Direction must be one of 'x', 'y', or 'z'"
    self.direction = direction
    resources = resources or []
    if direction == "z":  # top to bottom
      resources = list(reversed(resources))
    for resource in resources:
      self.assign_child_resource(resource)

  def get_size_x(self) -> float:
    """Get local size in the x direction."""
    if len(self.children) == 0:
      return 0
    if self.direction == "x":
      return sum(child.get_size_x() for child in self.children)
    return max(resource.get_size_x() for resource in self.children)

  def get_size_y(self) -> float:
    """Get local size in the y direction."""
    if len(self.children) == 0:
      return 0
    if self.direction == "y":
      return sum(child.get_size_y() for child in self.children)
    return max(resource.get_size_y() for resource in self.children)

  def get_size_z(self) -> float:
    """Get local size in the z direction."""

    def get_actual_resource_height(resource: Resource) -> float:
      """Helper function to get the actual height of a resource, accounting for the lid nesting
      height if the resource is a plate with a lid."""
      if isinstance(resource, Plate) and resource.lid is not None:
        return resource.get_size_z() + resource.lid.get_size_z() - resource.lid.nesting_z_height
      return resource.get_size_z()

    if len(self.children) == 0:
      return 0

    if self.direction != "z":
      return max(get_actual_resource_height(child) for child in self.children)
    return sum(get_actual_resource_height(child) for child in self.children)

  def get_resource_stack_edge(self) -> Coordinate:
    if self.direction == "x":
      resource_location = Coordinate(self.get_size_x(), 0, 0)
    elif self.direction == "y":
      resource_location = Coordinate(0, self.get_size_y(), 0)
    elif self.direction == "z":
      resource_location = Coordinate(0, 0, self.get_size_z())
    else:
      raise ValueError("self.direction must be one of 'x', 'y', or 'z'")

    return resource_location

  def get_new_child_location(self, resource: Resource) -> Coordinate:
    """Get the location where a new child resource should be placed in the stack."""
    return get_child_location(resource) + self.get_resource_stack_edge()

  def assign_child_resource(
    self, resource: Resource, location: Optional[Coordinate] = None, reassign: bool = True
  ):
    location = location or self.get_new_child_location(resource)
    return super().assign_child_resource(resource=resource, location=location, reassign=reassign)

  def unassign_child_resource(self, resource: Resource):
    if self.direction == "z" and resource != self.children[-1]:  # no floating resources
      raise ValueError("Resource is not the top item in this z-growing stack, cannot unassign")
    return super().unassign_child_resource(resource)

  def get_top_item(self) -> Resource:
    """Get the top item in the stack.

    For stacks growing in the x, y or z direction, this is the rightmost, frontmost, or topmost
    item in the stack, respectively.

    Returns:
      The top item in the stack.

    Raises:
      ValueError: If the stack is empty.
    """

    if len(self.children) == 0:
      raise ValueError("Stack is empty")

    return self.children[-1]

  def check_can_drop_resource_here(self, resource: Resource, *, reassign: bool = True) -> None:
    # for now, any resource can be dropped onto a stack.
    pass
